@@ -27,7 +27,7 @@
 
 #include "operations.h"
 #include "bitutils.h"
-#include "target.h"
+#include "hwunit.h"
 #include "debug.h"
 
 #define HTM_ERR(x) ({int rc = (x); if (rc) {PR_ERROR("HTM Error %d %s:%d\n", \
@@ -171,11 +171,13 @@ struct htm_status {
 
 static struct htm *check_and_convert(struct pdbg_target *target)
 {
-
 	if (!pdbg_target_is_class(target, "nhtm") &&
 	    !pdbg_target_is_class(target, "chtm"))
 	    return NULL;
 
+	if (pdbg_target_status(target) != PDBG_TARGET_ENABLED)
+		return NULL;
+
 	return target_to_htm(target);
 }
 
@@ -183,21 +185,45 @@ int htm_start(struct pdbg_target *target)
 {
 	struct htm *htm = check_and_convert(target);
 
-	return htm ? htm->start(htm) : -1;
+	if (!htm)
+		return -1;
+
+	if (!htm->start) {
+		PR_ERROR("start() not implemented for the target\n");
+		return -1;
+	}
+
+	return htm->start(htm);
 }
 
 int htm_stop(struct pdbg_target *target)
 {
 	struct htm *htm = check_and_convert(target);
 
-	return htm ? htm->stop(htm) : -1;
+	if (!htm)
+		return -1;
+
+	if (!htm->stop) {
+		PR_ERROR("stop() not implemented for the target\n");
+		return -1;
+	}
+
+	return htm->stop(htm);
 }
 
 int htm_status(struct pdbg_target *target)
 {
 	struct htm *htm = check_and_convert(target);
 
-	return htm ? htm->status(htm) : -1;
+	if (!htm)
+		return -1;
+
+	if (!htm->status) {
+		PR_ERROR("status() not implemented for the target\n");
+		return -1;
+	}
+
+	return htm->status(htm);
 }
 
 int htm_dump(struct pdbg_target *target, char *filename)
@@ -207,6 +233,11 @@ int htm_dump(struct pdbg_target *target, char *filename)
 	if (!htm || !filename)
 		return -1;
 
+	if (!htm->dump) {
+		PR_ERROR("dump() not implemented for the target\n");
+		return -1;
+	}
+
 	return htm->dump(htm, filename);
 }
 
@@ -217,6 +248,11 @@ int htm_record(struct pdbg_target *target, char *filename)
 	if (!htm || !filename)
 		return -1;
 
+	if (!htm->record) {
+		PR_ERROR("record() not implemented for the target\n");
+		return -1;
+	}
+
 	return htm->record(htm, filename);
 }
 
@@ -520,7 +556,7 @@ static int configure_nhtm(struct htm *htm, bool wrap)
 			NHTM_TTYPE_SIZE_MASK ))) /* no pattern matching */
 		return -1;
 
-	if (dt_node_is_compatible(&htm->target, "ibm,power9-nhtm")) {
+	if (pdbg_target_compatible(&htm->target, "ibm,power9-nhtm")) {
 		if (HTM_ERR(pib_read(&htm->target, NHTM_FLEX_MUX, &val)))
 			return -1;
 
@@ -548,18 +584,16 @@ static int is_startable(struct htm_status *status)
 
 static char *get_debugfs_file(struct htm *htm, const char *file)
 {
-	uint32_t chip_id;
+	struct pdbg_target *pib;
+	uint32_t index;
 	char *filename;
 
-	chip_id = dt_get_chip_id(&htm->target);
-	if (chip_id == -1) {
-		PR_ERROR("Couldn't find a chip id\n");
-		return NULL;
-	}
-
-	if (asprintf(&filename, "%s/%08x/%s", DEBUGFS_MEMTRACE, chip_id, file) == -1) {
+	pib = pdbg_target_parent("pib", &htm->target);
+	assert(pib);
+	index = pdbg_target_index(pib);
+	if (asprintf(&filename, "%s/%08x/%s", DEBUGFS_MEMTRACE, index, file) == -1) {
 		PR_ERROR("Couldn't asprintf() '%s/%08x/size': %m\n",
-				DEBUGFS_MEMTRACE, chip_id);
+				DEBUGFS_MEMTRACE, index);
 		return NULL;
 	}
 
@@ -740,7 +774,7 @@ static int htm_toggle_debug_bit(struct htm *htm)
 		return 0; /* nhtm case */
 
 	/* FIXME: this is a hack for P8 */
-	if (!dt_node_is_compatible(core, "ibm,power8-core")) {
+	if (!pdbg_target_compatible(core, "ibm,power8-core")) {
 		PR_ERROR("HTM is POWER8 only currently\n");
 		return -1;
 	}
@@ -784,7 +818,7 @@ static int __do_htm_start(struct htm *htm, bool wrap)
 	/*
 	 * Instead of the HTM_TRIG_START, this is where you might want
 	 * to call do_adu_magic()
-	 * for_each_child_target("adu", core, do_adu_magic, NULL, NULL);
+	 * for_each_child_target("mem", core, do_adu_magic, NULL, NULL);
 	 * see what I mean?
 	 */
 
@@ -867,7 +901,7 @@ static int do_htm_status(struct htm *htm)
 	uint64_t val, total;
 	int i, regs = 9;
 
-	if (dt_node_is_compatible(&htm->target, "ibm,power9-nhtm"))
+	if (pdbg_target_compatible(&htm->target, "ibm,power9-nhtm"))
 		regs++;
 
 	PR_INFO("HTM register dump:\n");
@@ -975,7 +1009,6 @@ static int do_htm_dump(struct htm *htm, char *filename)
 	uint64_t last, end, trace_size;
 	int trace_fd, dump_fd;
 	uint32_t eyecatcher;
-	uint32_t chip_id;
 	size_t r;
 	bool wrapped;
 
@@ -990,10 +1023,6 @@ static int do_htm_dump(struct htm *htm, char *filename)
 		return -1;
 	}
 
-	chip_id = dt_get_chip_id(&htm->target);
-	if (chip_id == -1)
-		return -1;
-
 	trace_file = get_debugfs_file(htm, "trace");
 	if (!trace_file)
 		return -1;
@@ -1101,7 +1130,7 @@ static int nhtm_probe(struct pdbg_target *target)
 	if (!is_debugfs_memtrace_ok() || !is_debugfs_scom_ok())
 		return -1;
 
-	if (dt_node_is_compatible(target, "ibm,power9-nhtm")) {
+	if (pdbg_target_compatible(target, "ibm,power9-nhtm")) {
 		pib_read(target, NHTM_FLEX_MUX, &val);
 		if (GETFIELD(NHTM_FLEX_MUX_MASK, val) != NHTM_FLEX_DEFAULT) {
 			PR_DEBUG("FLEX_MUX not default\n");
@@ -1117,10 +1146,10 @@ static int chtm_probe(struct pdbg_target *target)
 	return is_debugfs_memtrace_ok() && is_debugfs_scom_ok() ? 0 : -1;
 }
 
-static struct htm nhtm = {
+static struct htm p8_nhtm = {
 	.target = {
-		.name =	"Nest HTM",
-		.compatible = "ibm,power8-nhtm", "ibm,power9-nhtm",
+		.name =	"POWER8 Nest HTM",
+		.compatible = "ibm,power8-nhtm",
 		.class = "nhtm",
 		.probe = nhtm_probe,
 	},
@@ -1130,9 +1159,24 @@ static struct htm nhtm = {
 	.status = do_htm_status,
 	.dump = do_htm_dump,
 };
-DECLARE_HW_UNIT(nhtm);
+DECLARE_HW_UNIT(p8_nhtm);
 
-static struct htm chtm = {
+static struct htm p9_nhtm = {
+	.target = {
+		.name =	"POWER9 Nest HTM",
+		.compatible = "ibm,power9-nhtm",
+		.class = "nhtm",
+		.probe = nhtm_probe,
+	},
+	.start = do_htm_start,
+	.stop = do_htm_stop,
+	.record = do_htm_record,
+	.status = do_htm_status,
+	.dump = do_htm_dump,
+};
+DECLARE_HW_UNIT(p9_nhtm);
+
+static struct htm p8_chtm = {
 	.target = {
 		.name = "POWER8 Core HTM",
 		.compatible = "ibm,power8-chtm",
@@ -1145,4 +1189,12 @@ static struct htm chtm = {
 	.status = do_htm_status,
 	.dump = do_htm_dump,
 };
-DECLARE_HW_UNIT(chtm);
+DECLARE_HW_UNIT(p8_chtm);
+
+__attribute__((constructor))
+static void register_htm(void)
+{
+	pdbg_hwunit_register(PDBG_DEFAULT_BACKEND, &p8_nhtm_hw_unit);
+	pdbg_hwunit_register(PDBG_DEFAULT_BACKEND, &p9_nhtm_hw_unit);
+	pdbg_hwunit_register(PDBG_DEFAULT_BACKEND, &p8_chtm_hw_unit);
+}
