@@ -31,37 +31,20 @@
 #include <config.h>
 
 #include <libpdbg.h>
-#include <target.h>
 
-#include "main.h"
 #include "htm.h"
-#include "options.h"
 #include "optcmd.h"
 #include "progress.h"
+#include "pdbgproxy.h"
+#include "util.h"
+#include "path.h"
 
 #define PR_ERROR(x, args...) \
 	pdbg_log(PDBG_ERROR, x, ##args)
 
-#include "fake.dt.h"
-
-#ifdef TARGET_ARM
-#include "p8-i2c.dt.h"
-#include "p8-fsi.dt.h"
-#include "p8-kernel.dt.h"
-#include "p9w-fsi.dt.h"
-#include "p9r-fsi.dt.h"
-#include "p9z-fsi.dt.h"
-#include "p9-kernel.dt.h"
-#endif
-
-#ifdef TARGET_PPC
-#include "p8-host.dt.h"
-#include "p9-host.dt.h"
-#endif
-
 #define THREADS_PER_CORE	8
 
-static enum backend backend = KERNEL;
+static enum pdbg_backend backend = PDBG_DEFAULT_BACKEND;
 
 static char const *device_node;
 static int i2c_addr = 0x50;
@@ -72,9 +55,10 @@ static int i2c_addr = 0x50;
 
 #define MAX_LINUX_CPUS	(MAX_PROCESSORS * MAX_CHIPS * MAX_THREADS)
 
-static int **processorsel[MAX_PROCESSORS];
-static int *chipsel[MAX_PROCESSORS][MAX_CHIPS];
-static int threadsel[MAX_PROCESSORS][MAX_CHIPS][MAX_THREADS];
+#define MAX_PATH_ARGS	16
+
+static const char *pathsel[MAX_PATH_ARGS];
+static int pathsel_count;
 
 static int probe(void);
 
@@ -87,8 +71,10 @@ extern struct optcmd_cmd
 	optcmd_getnia, optcmd_putnia, optcmd_getmsr, optcmd_putmsr,
 	optcmd_getring, optcmd_start, optcmd_stop, optcmd_step,
 	optcmd_threadstatus, optcmd_sreset, optcmd_regs, optcmd_probe,
-	optcmd_getmem, optcmd_putmem, optcmd_getxer, optcmd_putxer,
-	optcmd_getcr, optcmd_putcr;
+	optcmd_getmem, optcmd_putmem, optcmd_getmemio, optcmd_putmemio,
+	optcmd_getmempba, optcmd_putmempba,
+	optcmd_getxer, optcmd_putxer, optcmd_getcr, optcmd_putcr,
+	optcmd_gdbserver, optcmd_istep;
 
 static struct optcmd_cmd *cmds[] = {
 	&optcmd_getscom, &optcmd_putscom, &optcmd_getcfam, &optcmd_putcfam,
@@ -96,8 +82,10 @@ static struct optcmd_cmd *cmds[] = {
 	&optcmd_getnia, &optcmd_putnia, &optcmd_getmsr, &optcmd_putmsr,
 	&optcmd_getring, &optcmd_start, &optcmd_stop, &optcmd_step,
 	&optcmd_threadstatus, &optcmd_sreset, &optcmd_regs, &optcmd_probe,
-	&optcmd_getmem, &optcmd_putmem, &optcmd_getxer, &optcmd_putxer,
-	&optcmd_getcr, &optcmd_putcr,
+	&optcmd_getmem, &optcmd_putmem, &optcmd_getmemio, &optcmd_putmemio,
+	&optcmd_getmempba, &optcmd_putmempba,
+	&optcmd_getxer, &optcmd_putxer, &optcmd_getcr, &optcmd_putcr,
+	&optcmd_gdbserver, &optcmd_istep,
 };
 
 /* Purely for printing usage text. We could integrate printing argument and flag
@@ -131,18 +119,24 @@ static struct action actions[] = {
 	{ "putcfam", "<address> <value> [<mask>]", "Write system cfam" },
 	{ "getscom", "<address>", "Read system scom" },
 	{ "putscom", "<address> <value> [<mask>]", "Write system scom" },
-	{ "getmem",  "<address> <count>", "Read system memory" },
-	{ "putmem",  "<address>", "Write to system memory" },
+	{ "getmem",  "<address> <count> [--ci] [--raw]", "Read system memory" },
+	{ "getmempba",  "<address> <count> [--ci] [--raw]", "Read system memory" },
+	{ "getmemio", "<address> <count> <block size> [--raw]", "Read memory cache inhibited with specified transfer size" },
+	{ "putmem",  "<address> [--ci]", "Write to system memory" },
+	{ "putmempba",  "<address> [--ci]", "Write to system memory" },
+	{ "putmemio", "<address> <block size>", "Write system memory cache inhibited with specified transfer size" },
 	{ "threadstatus", "", "Print the status of a thread" },
 	{ "sreset",  "", "Reset" },
 	{ "regs",  "[--backtrace]", "State (optionally display backtrace)" },
+	{ "gdbserver", "", "Start a gdb server" },
+	{ "istep", "<major> <minor>|0", "Execute istep on SBE" },
 };
 
-static void print_usage(char *pname)
+static void print_usage(void)
 {
 	int i;
 
-	printf("Usage: %s [options] command ...\n\n", pname);
+	printf("Usage: pdbg [options] command ...\n\n");
 	printf(" Options:\n");
 	printf("\t-p, --processor=<0-%d>|<range>|<list>\n", MAX_PROCESSORS-1);
 	printf("\t-c, --chip=<0-%d>|<range>|<list>\n", MAX_CHIPS-1);
@@ -150,9 +144,12 @@ static void print_usage(char *pname)
 #ifdef TARGET_PPC
 	printf("\t-l, --cpu=<0-%d>|<range>|<list>\n", MAX_PROCESSORS-1);
 #endif
+	printf("\t-P, --path=<device tree node spec>\n");
 	printf("\t-a, --all\n");
 	printf("\t\tRun command on all possible processors/chips/threads (default)\n");
 	printf("\t-b, --backend=backend\n");
+	printf("\t\tcronus:\tA backend based on cronus server\n");
+	printf("\t\tsbefifo:\tA backend using sbefifo kernel driver\n");
 	printf("\t\tfsi:\tAn experimental backend that uses\n");
 	printf("\t\t\tbit-banging to access the host processor\n");
 	printf("\t\t\tvia the FSI bus.\n");
@@ -178,79 +175,6 @@ static void print_usage(char *pname)
 		printf("  %-15s %-27s  %s\n", actions[i].name, actions[i].args, actions[i].desc);
 }
 
-/* Parse argument of the form 0-5,7,9-11,15,17 */
-static bool parse_list(const char *arg, int max, int *list, int *count)
-{
-	char str[strlen(arg)+1];
-	char *tok, *tmp, *saveptr = NULL;
-	int i;
-
-	assert(max < INT_MAX);
-
-	strcpy(str, arg);
-
-	tmp = str;
-	while ((tok = strtok_r(tmp, ",", &saveptr)) != NULL) {
-		char *a, *b, *endptr, *saveptr2 = NULL;
-		unsigned long int from, to;
-
-		a = strtok_r(tok, "-", &saveptr2);
-		if (a == NULL) {
-			return false;
-		} else {
-			endptr = NULL;
-			from = strtoul(a, &endptr, 0);
-			if (*endptr != '\0') {
-				fprintf(stderr, "Invalid value %s\n", a);
-				return false;
-			}
-			if (from >= max) {
-				fprintf(stderr, "Value %s larger than max %d\n", a, max-1);
-				return false;
-			}
-		}
-
-		b = strtok_r(NULL, "-", &saveptr2);
-		if (b == NULL) {
-			to = from;
-		} else {
-			endptr = NULL;
-			to = strtoul(b, &endptr, 0);
-			if (*endptr != '\0') {
-				fprintf(stderr, "Invalid value %s\n", b);
-				return false;
-			}
-			if (to >= max) {
-				fprintf(stderr, "Value %s larger than max %d\n", b, max-1);
-				return false;
-			}
-		}
-
-		if (from > to) {
-			fprintf(stderr, "Invalid range %s-%s\n", a, b);
-			return false;
-		}
-
-		for (i = from; i <= to; i++)
-			list[i] = 1;
-
-		tmp = NULL;
-	};
-
-	if (count != NULL) {
-		int n = 0;
-
-		for (i = 0; i < max; i++) {
-			if (list[i] == 1)
-				n++;
-		}
-
-		*count = n;
-	}
-
-	return true;
-}
-
 #ifdef TARGET_PPC
 int get_pir(int linux_cpu)
 {
@@ -313,6 +237,56 @@ void pir_map(int pir, int *chip, int *core, int *thread) {}
 #define PPC_OPTS
 #endif
 
+static bool pathsel_add(char *format, ...) __attribute__((format (printf, 1, 2)));
+
+static bool pathsel_add(char *format, ...)
+{
+	va_list ap;
+	char path[1024];
+	int len;
+
+	va_start(ap, format);
+
+	len = vsnprintf(path, sizeof(path), format, ap);
+	if (len > sizeof(path)) {
+		va_end(ap);
+		return false;
+	}
+
+	va_end(ap);
+
+	if (pathsel_count == MAX_PATH_ARGS) {
+		fprintf(stderr, "Too many path arguments\n");
+		return false;
+	}
+
+	pathsel[pathsel_count] = strdup(path);
+	assert(pathsel[pathsel_count]);
+	pathsel_count++;
+
+	return true;
+}
+
+static bool list_to_string(int *list, int max, char *str, size_t len)
+{
+	char tmp[16];
+	int i;
+
+	memset(str, 0, len);
+
+	for (i=0; i<max; i++) {
+		if (list[i] == 1) {
+			sprintf(tmp, "%d,", i);
+			if (strlen(str) + strlen(tmp) + 1 > len) {
+				return false;
+			}
+			strcat(str, tmp);
+		}
+	}
+
+	return true;
+}
+
 static bool parse_options(int argc, char *argv[])
 {
 	int c;
@@ -322,7 +296,7 @@ static bool parse_options(int argc, char *argv[])
 	int t_list[MAX_THREADS];
 	int l_list[MAX_LINUX_CPUS];
 	int p_count = 0, c_count = 0, t_count = 0, l_count = 0;
-	int i, j, k;
+	int i;
 	struct option long_opts[] = {
 		{"all",			no_argument,		NULL,	'a'},
 		{"backend",		required_argument,	NULL,	'b'},
@@ -336,11 +310,13 @@ static bool parse_options(int argc, char *argv[])
 		{"cpu",			required_argument,	NULL,	'l'},
 #endif
 		{"debug",		required_argument,	NULL,	'D'},
+		{"path",		required_argument,	NULL,	'P'},
 		{"shutup",		no_argument,		NULL,	'S'},
 		{"version",		no_argument,		NULL,	'V'},
 		{NULL,			0,			NULL,     0}
 	};
 	char *endptr;
+	char p_str[256], c_str[256], t_str[256];
 
 	memset(p_list, 0, sizeof(p_list));
 	memset(c_list, 0, sizeof(c_list));
@@ -348,7 +324,7 @@ static bool parse_options(int argc, char *argv[])
 	memset(l_list, 0, sizeof(l_list));
 
 	do {
-		c = getopt_long(argc, argv, "+ab:c:d:hp:s:t:D:SV" PPC_OPTS,
+		c = getopt_long(argc, argv, "+ab:c:d:hp:s:t:D:P:SV" PPC_OPTS,
 				long_opts, NULL);
 		if (c == -1)
 			break;
@@ -404,20 +380,23 @@ static bool parse_options(int argc, char *argv[])
 
 		case 'b':
 			if (strcmp(optarg, "fsi") == 0) {
-				backend = FSI;
+				backend = PDBG_BACKEND_FSI;
 			} else if (strcmp(optarg, "i2c") == 0) {
-				backend = I2C;
+				backend = PDBG_BACKEND_I2C;
 			} else if (strcmp(optarg, "kernel") == 0) {
-				backend = KERNEL;
+				backend = PDBG_BACKEND_KERNEL;
 				/* TODO: use device node to point at a slave
 				 * other than the first? */
 			} else if (strcmp(optarg, "fake") == 0) {
-				backend = FAKE;
+				backend = PDBG_BACKEND_FAKE;
 			} else if (strcmp(optarg, "host") == 0) {
-				backend = HOST;
+				backend = PDBG_BACKEND_HOST;
+			} else if (strcmp(optarg, "cronus") == 0) {
+				backend = PDBG_BACKEND_CRONUS;
+			} else if (strcmp(optarg, "sbefifo") == 0) {
+				backend = PDBG_BACKEND_SBEFIFO;
 			} else {
 				fprintf(stderr, "Invalid backend '%s'\n", optarg);
-				print_backends(stderr);
 				opt_error = true;
 			}
 			break;
@@ -434,6 +413,11 @@ static bool parse_options(int argc, char *argv[])
 				fprintf(stderr, "Invalid slave address '%s'\n", optarg);
 			break;
 
+		case 'P':
+			if (!pathsel_add("%s", optarg))
+				opt_error = true;
+			break;
+
 		case 'S':
 			progress_shutup();
 			break;
@@ -450,7 +434,7 @@ static bool parse_options(int argc, char *argv[])
 		case '?':
 		case 'h':
 			opt_error = true;
-			print_usage(basename(argv[0]));
+			print_usage();
 			break;
 		}
 	} while (c != EOF && !opt_error);
@@ -464,6 +448,16 @@ static bool parse_options(int argc, char *argv[])
 		return false;
 	}
 
+	if (pathsel_count > 0 && l_count > 0) {
+		fprintf(stderr, "Can't mix -l with -P\n");
+		return false;
+	}
+
+	if ((c_count > 0 || t_count > 0 || p_count > 0) && (pathsel_count > 0)) {
+		fprintf(stderr, "Can't mix -P with -p/-c/-t/-a\n");
+		return false;
+	}
+
 	if ((c_count > 0 || t_count > 0) && p_count == 0) {
 		fprintf(stderr, "No processor(s) selected\n");
 		fprintf(stderr, "Use -p or -a to select processor(s)\n");
@@ -476,30 +470,35 @@ static bool parse_options(int argc, char *argv[])
 		return false;
 	}
 
-	for (i = 0; i < MAX_PROCESSORS; i++) {
-		if (p_list[i] == 0)
-			continue;
+	if (p_count) {
+		if (!list_to_string(p_list, MAX_PROCESSORS, p_str, sizeof(p_str)))
+			return false;
+		if (!pathsel_add("fsi[%s]", p_str))
+			return false;
+		if (!pathsel_add("pib[%s]", p_str))
+			return false;
+	}
 
-		processorsel[i] = &chipsel[i][0];
+	if (c_count) {
+		if (!list_to_string(c_list, MAX_CHIPS, c_str, sizeof(c_str)))
+			return false;
+		if (!pathsel_add("pib[%s]/core[%s]", p_str, c_str))
+			return false;
+	}
 
-		for (j = 0; j < MAX_CHIPS; j++) {
-			if (c_list[j] == 0)
-				continue;
-
-			chipsel[i][j] = &threadsel[i][j][0];
-
-			for (k = 0; k < MAX_THREADS; k++) {
-				if (t_list[k] == 0)
-					continue;
-
-				threadsel[i][j][k] = 1;
-			}
-		}
+	if (t_count) {
+		if (!list_to_string(t_list, MAX_THREADS, t_str, sizeof(t_str)))
+			return false;
+		if (!pathsel_add("pib[%s]/core[%s]/thread[%s]", p_str, c_str, t_str))
+			return false;
 	}
 
 	if (l_count) {
 		int pir = -1, i, chip, core, thread;
 
+		if (!device_node)
+			return false;
+
 		for (i = 0; i < MAX_LINUX_CPUS; i++) {
 			if (l_list[i] == 1) {
 				pir = get_pir(i);
@@ -508,9 +507,12 @@ static bool parse_options(int argc, char *argv[])
 
 				pir_map(pir, &chip, &core, &thread);
 
-				processorsel[chip] = &chipsel[chip][0];
-				chipsel[chip][core] = &threadsel[chip][core][0];
-				threadsel[chip][core][thread] = 1;
+				if (!pathsel_add("pib%d", chip))
+					return false;
+				if (!pathsel_add("pib%d/core%d", chip, core))
+					return false;
+				if (!pathsel_add("pib%d/core%d/thread%d", chip, core, thread))
+					return false;
 			}
 		}
 	}
@@ -518,218 +520,12 @@ static bool parse_options(int argc, char *argv[])
 	return true;
 }
 
-void target_select(struct pdbg_target *target)
-{
-	/* We abuse the private data pointer atm to indicate the target is
-	 * selected */
-	pdbg_target_priv_set(target, (void *) 1);
-}
-
-void target_unselect(struct pdbg_target *target)
-{
-	pdbg_target_priv_set(target, NULL);
-}
-
-bool target_selected(struct pdbg_target *target)
-{
-	return (bool) pdbg_target_priv(target);
-}
-
-/* Returns the sum of return codes. This can be used to count how many targets the callback was run on. */
-int for_each_child_target(char *class, struct pdbg_target *parent,
-				 int (*cb)(struct pdbg_target *, uint32_t, uint64_t *, uint64_t *),
-				 uint64_t *arg1, uint64_t *arg2)
-{
-	int rc = 0;
-	struct pdbg_target *target;
-	uint32_t index;
-	enum pdbg_target_status status;
-
-	pdbg_for_each_target(class, parent, target) {
-		if (!target_selected(target))
-			continue;
-
-		index = pdbg_target_index(target);
-		assert(index != -1);
-		pdbg_target_probe(target);
-		status = pdbg_target_status(target);
-		if (status != PDBG_TARGET_ENABLED)
-			continue;
-
-		rc += cb(target, index, arg1, arg2);
-	}
-
-	return rc;
-}
-
-int for_each_target(char *class, int (*cb)(struct pdbg_target *, uint32_t, uint64_t *, uint64_t *), uint64_t *arg1, uint64_t *arg2)
-{
-	struct pdbg_target *target;
-	uint32_t index;
-	enum pdbg_target_status status;
-	int rc = 0;
-
-	pdbg_for_each_class_target(class, target) {
-		if (!target_selected(target))
-			continue;
-
-		index = pdbg_target_index(target);
-		assert(index != -1);
-		pdbg_target_probe(target);
-		status = pdbg_target_status(target);
-		if (status != PDBG_TARGET_ENABLED)
-			continue;
-
-		rc += cb(target, index, arg1, arg2);
-	}
-
-	return rc;
-}
-
-void for_each_target_release(char *class)
-{
-	struct pdbg_target *target;
-
-	pdbg_for_each_class_target(class, target) {
-		if (!target_selected(target))
-			continue;
-
-		pdbg_target_release(target);
-	}
-}
-
-static int target_selection(void)
-{
-	struct pdbg_target *fsi, *pib, *chip, *thread;
-
-	switch (backend) {
-#ifdef TARGET_ARM
-	case I2C:
-		pdbg_targets_init(&_binary_p8_i2c_dtb_o_start);
-		break;
-
-	case FSI:
-		if (device_node == NULL) {
-			PR_ERROR("FSI backend requires a device type\n");
-			return -1;
-		}
-		if (!strcmp(device_node, "p8"))
-			pdbg_targets_init(&_binary_p8_fsi_dtb_o_start);
-		else if (!strcmp(device_node, "p9w"))
-			pdbg_targets_init(&_binary_p9w_fsi_dtb_o_start);
-		else if (!strcmp(device_node, "p9r"))
-			pdbg_targets_init(&_binary_p9r_fsi_dtb_o_start);
-		else if (!strcmp(device_node, "p9z"))
-			pdbg_targets_init(&_binary_p9z_fsi_dtb_o_start);
-		else {
-			PR_ERROR("Invalid device type specified\n");
-			return -1;
-		}
-		break;
-
-	case KERNEL:
-		if (device_node == NULL) {
-			PR_ERROR("kernel backend requires a device type\n");
-                        return -1;
-		}
-		if (!strcmp(device_node, "p8"))
-			pdbg_targets_init(&_binary_p8_kernel_dtb_o_start);
-		else
-			pdbg_targets_init(&_binary_p9_kernel_dtb_o_start);
-		break;
-
-#endif
-
-#ifdef TARGET_PPC
-	case HOST:
-		if (device_node == NULL) {
-			PR_ERROR("Host backend requires a device type\n");
-			return -1;
-		}
-		if (!strcmp(device_node, "p8"))
-			pdbg_targets_init(&_binary_p8_host_dtb_o_start);
-		else if (!strcmp(device_node, "p9"))
-			pdbg_targets_init(&_binary_p9_host_dtb_o_start);
-		else {
-			PR_ERROR("Unsupported device type for host backend\n");
-			return -1;
-		}
-		break;
-#endif
-
-	case FAKE:
-		pdbg_targets_init(&_binary_fake_dtb_o_start);
-		break;
-
-	default:
-		/* parse_options deals with parsing user input, so it should be
-		 * impossible to get here */
-		assert(0);
-		return -1;
-	}
-
-	/* At this point we should have a device-tree loaded. We want
-	 * to walk the tree and disabled nodes we don't care about
-	 * prior to probing. */
-	pdbg_for_each_class_target("pib", pib) {
-		int proc_index = pdbg_target_index(pib);
-
-		if (backend == I2C && device_node)
-			pdbg_set_target_property(pib, "bus", device_node, strlen(device_node) + 1);
-
-		if (processorsel[proc_index]) {
-			target_select(pib);
-			pdbg_for_each_target("core", pib, chip) {
-				int chip_index = pdbg_target_index(chip);
-				if (pdbg_parent_index(chip, "pib") != proc_index)
-					continue;
-
-				if (chipsel[proc_index][chip_index]) {
-					target_select(chip);
-					pdbg_for_each_target("thread", chip, thread) {
-						int thread_index = pdbg_target_index(thread);
-						if (threadsel[proc_index][chip_index][thread_index])
-							target_select(thread);
-						else
-							target_unselect(thread);
-					}
-				} else
-					target_unselect(chip);
-			}
-
-			/* This is kinda broken as we're overloading what '-c'
-			 * means - it's now up to each command to select targets
-			 * based on core/chiplet. We really need a better
-			 * solution to target selection. */
-			pdbg_for_each_target("chiplet", pib, chip) {
-				int chip_index = pdbg_target_index(chip);
-				if (chipsel[proc_index][chip_index]) {
-					target_select(chip);
-				} else
-					target_unselect(chip);
-			}
-		} else
-			target_unselect(pib);
-	}
-
-	pdbg_for_each_class_target("fsi", fsi) {
-		int index = pdbg_target_index(fsi);
-		if (processorsel[index])
-			target_select(fsi);
-		else
-			target_unselect(fsi);
-	}
-
-	return 0;
-}
-
-void print_target(struct pdbg_target *target, int level)
+static void print_target(struct pdbg_target *target, int level)
 {
 	int i;
 	struct pdbg_target *next;
 	enum pdbg_target_status status;
-
-	pdbg_target_probe(target);
+	const char *classname;
 
 	/* Does this target actually exist? */
 	status = pdbg_target_status(target);
@@ -739,29 +535,16 @@ void print_target(struct pdbg_target *target, int level)
 	for (i = 0; i < level; i++)
 		printf("    ");
 
-	if (target) {
-		char c = 0;
-
-		if (!pdbg_target_class_name(target))
-			return;
-
-		if (!strcmp(pdbg_target_class_name(target), "pib"))
-			c = 'p';
-		else if (!strcmp(pdbg_target_class_name(target), "core"))
-			c = 'c';
-		else if (!strcmp(pdbg_target_class_name(target), "thread"))
-			c = 't';
-
-		if (c)
-			printf("%c%d: %s\n", c, pdbg_target_index(target), pdbg_target_name(target));
-		else
-			printf("%s\n", pdbg_target_name(target));
-	}
+	classname = pdbg_target_class_name(target);
+	if (classname)
+		printf("%s%d: %s", classname, pdbg_target_index(target), pdbg_target_name(target));
+	else
+		printf("%s:", pdbg_target_dn_name(target));
+	if (path_target_selected(target))
+		printf(" (*)");
+	printf("\n");
 
 	pdbg_for_each_child_target(target, next) {
-		if (!target_selected(next))
-			continue;
-
 		print_target(next, level + 1);
 	}
 }
@@ -770,16 +553,10 @@ static int probe(void)
 {
 	struct pdbg_target *target;
 
-	pdbg_for_each_class_target("pib", target) {
-		if (!target_selected(target))
-			continue;
-
+	pdbg_for_each_child_target(pdbg_target_root(), target) {
 		print_target(target, 0);
 	}
 
-	printf("\nNote that only selected targets will be shown above. If none are shown\n"
-			"try adding '-a' to select all targets\n");
-
 	return 1;
 }
 OPTCMD_DEFINE_CMD(probe, probe);
@@ -789,7 +566,7 @@ OPTCMD_DEFINE_CMD(probe, probe);
  */
 static void atexit_release(void)
 {
-	pdbg_target_release(dt_root);
+	pdbg_target_release(pdbg_target_root());
 }
 
 int main(int argc, char *argv[])
@@ -797,24 +574,39 @@ int main(int argc, char *argv[])
 	int i, rc = 0;
 	void **args, **flags;
 	optcmd_cmd_t *cmd;
-
-	backend = default_backend();
+	struct pdbg_target *target;
 
 	if (!parse_options(argc, argv))
 		return 1;
 
 	if (optind >= argc) {
-		print_usage(basename(argv[0]));
+		print_usage();
 		return 1;
 	}
 
-	if (!device_node)
-		device_node = default_target(backend);
+	if (backend)
+		if (!pdbg_set_backend(backend, device_node))
+			return 1;
 
-	/* Disable unselected targets */
-	if (target_selection())
+	if (!pdbg_targets_init(NULL))
 		return 1;
 
+	if (pathsel_count) {
+		if (!path_target_parse(pathsel, pathsel_count))
+			return 1;
+	}
+
+	if (!path_target_present()) {
+		printf("No valid targets found or specified. Try adding -p/-c/-t options to specify a target.\n");
+		printf("Alternatively run 'pdbg -a probe' to get a list of all valid targets\n");
+		return 1;
+	}
+
+	/* Probe all selected targets */
+	for_each_path_target(target) {
+		pdbg_target_probe(target);
+	}
+
 	atexit(atexit_release);
 
 	for (i = 0; i < ARRAY_SIZE(cmds); i++) {
@@ -846,8 +638,5 @@ found_action:
 	if (rc > 0)
 		return 0;
 
-	printf("No valid targets found or specified. Try adding -p/-c/-t options to specify a target.\n");
-	printf("Alternatively run '%s -a probe' to get a list of all valid targets\n",
-	       basename(argv[0]));
 	return 1;
 }
